1#ifndef AMREX_MF_PARALLEL_FOR_G_H_
2#define AMREX_MF_PARALLEL_FOR_G_H_
3#include <AMReX_Config.H>
14namespace amrex::detail {
17void build_par_for_boxes (
char*& hp,
BoxIndexer*& pboxes, Vector<Box>
const& boxes)
19 if (boxes.empty()) {
return; }
20 const int nboxes = boxes.size();
21 const std::size_t nbytes = nboxes*
sizeof(
BoxIndexer);
24 for (
int i = 0; i < nboxes; ++i) {
34void destroy_par_for_boxes (
char* hp,
char* dp)
46namespace parfor_mf_detail {
49 auto call_f (F
const& f,
int b,
int i,
int j,
int k,
int)
noexcept
50 ->
decltype(f(0,0,0,0))
57 auto call_f (F
const& f,
int b,
int i,
int j,
int k,
int ncomp)
noexcept
58 ->
decltype(f(0,0,0,0,0))
60 for (
int n = 0; n < ncomp; ++n) {
66template <
int MT, FabArrayType MF,
typename F>
68ParallelFor_doit (MF
const& mf,
IntVect const& nghost,
int ncomp,
IntVect const&,
bool,
F const& f)
70 const auto& index_array = mf.IndexArray();
71 const int nboxes = index_array.size();
75 }
else if (nboxes == 1) {
79 parfor_mf_detail::call_f(f, 0, i, j, k, ncomp);
82 auto const& parforinfo = mf.getParForInfo(nghost);
83 auto nblocks_per_box = parforinfo.getNBlocksPerBox(MT);
85 const int nblocks = nblocks_per_box * nboxes;
86 const BoxIndexer* dp_boxes = parforinfo.getBoxes();
88#if defined(AMREX_USE_CUDA) || defined(AMREX_USE_HIP)
90 amrex::launch_global<MT>
94 int ibox =
int(blockIdx.x) / nblocks_per_box;
95 auto icell = std::uint64_t(blockIdx.x-ibox*nblocks_per_box)*MT + threadIdx.x;
97#elif defined(AMREX_USE_SYCL)
102 int blockIdxx = item.get_group_linear_id();
103 int threadIdxx = item.get_local_linear_id();
104 int ibox =
int(blockIdxx) / nblocks_per_box;
105 auto icell = std::uint64_t(blockIdxx-ibox*nblocks_per_box)*MT + threadIdxx;
108 if (icell < indexer.numPts()) {
109 auto [i, j, k] = indexer(icell);
110 parfor_mf_detail::call_f(f, ibox, i, j, k, ncomp);
117template <FabArrayType MF,
typename F>
119ParallelFor_doit (MF
const& mf,
IntVect const& nghost,
int ncomp,
IntVect const& ts,
bool dynamic,
F&& f)
122 constexpr int MT = 128;
124 constexpr int MT = AMREX_GPU_MAX_THREADS;
126 ParallelFor_doit<MT>(mf, nghost, ncomp, ts, dynamic, std::forward<F>(f));
129template <FabArrayType MF,
typename F>
131ParallelFor_doit (MF
const& mf,
IntVect const& nghost,
IntVect const& ts,
bool dynamic,
F&& f)
134 constexpr int MT = 128;
136 constexpr int MT = AMREX_GPU_MAX_THREADS;
138 ParallelFor_doit<MT>(mf, nghost, 1, ts, dynamic, std::forward<F>(f));
143template <
int MT, FabArrayType MF,
typename F>
147 const int nboxes = mf.IndexArray().size();
148 if (nboxes == 0) {
return; }
150 auto const& parforinfo = mf.getParForInfo(stride,
offset);
151 const auto nblocks_per_box = parforinfo.getNBlocksPerBox(MT);
152 if (nblocks_per_box == 0) {
return; }
154 const int nblocks = nblocks_per_box * nboxes;
155 const BoxIndexer* dp_boxes = parforinfo.getBoxes();
157#if defined(AMREX_USE_CUDA) || defined(AMREX_USE_HIP)
159 amrex::launch_global<MT>
163 int ibox =
int(blockIdx.x) / nblocks_per_box;
164 auto icell = std::uint64_t(blockIdx.x-ibox*nblocks_per_box)*MT + threadIdx.x;
166#elif defined(AMREX_USE_SYCL)
171 int blockIdxx = item.get_group_linear_id();
172 int threadIdxx = item.get_local_linear_id();
173 int ibox =
int(blockIdxx) / nblocks_per_box;
174 auto icell = std::uint64_t(blockIdxx-ibox*nblocks_per_box)*MT + threadIdxx;
177 if (icell < indexer.numPts()) {
178 IntVect iv = indexer.intVect(icell);
179 for (
int idim = 0; idim < AMREX_SPACEDIM; ++idim) {
180 iv[idim] = indexer.lo[idim] + (iv[idim]-indexer.lo[idim])*stride[idim];
182 auto const d3 = iv.
dim3();
183 f(ibox, d3.x, d3.y, d3.z);
191int box_indexer_len_x (
BoxIndexer const& indexer)
193#if (AMREX_SPACEDIM == 1)
194 return int(indexer.npts);
196 return int(indexer.fdm[0].divisor);
202template <
int MT, FabArrayType MF,
typename F>
204ParallelForRedBlack_doit (MF
const& mf,
int redblack,
F const& f)
206 const int nboxes = mf.IndexArray().size();
207 if (nboxes == 0) {
return; }
209 auto const& parforinfo = mf.getParForInfoRedBlack();
210 const auto nblocks_per_box = parforinfo.getNBlocksPerBox(MT);
211 if (nblocks_per_box == 0) {
return; }
213 const int nblocks = nblocks_per_box * nboxes;
214 const BoxIndexer* dp_boxes = parforinfo.getBoxes();
216#if defined(AMREX_USE_CUDA) || defined(AMREX_USE_HIP)
218 amrex::launch_global<MT>
222 int ibox =
int(blockIdx.x) / nblocks_per_box;
223 auto icell = std::uint64_t(blockIdx.x-ibox*nblocks_per_box)*MT + threadIdx.x;
225#elif defined(AMREX_USE_SYCL)
230 int blockIdxx = item.get_group_linear_id();
231 int threadIdxx = item.get_local_linear_id();
232 int ibox =
int(blockIdxx) / nblocks_per_box;
233 auto icell = std::uint64_t(blockIdxx-ibox*nblocks_per_box)*MT + threadIdxx;
237 BoxIndexer const* indexers = dp_boxes + 2*ibox;
238 const int big = (indexers[1].numPts() > indexers[0].numPts()) ? 1 : 0;
239 if (icell < indexers[big].numPts()) {
240 const IntVect iv = indexers[big].intVect(icell);
241 const int ic = iv[0] - indexers[big].lo[0];
242 auto const d3 = iv.
dim3();
243 const int parity = (d3.y + d3.z + redblack) & 1;
245 if (ic < box_indexer_len_x(indexer)) {
246 f(ibox, indexer.lo[0] + 2*ic, d3.y, d3.z);
253template <FabArrayType MF,
typename F>
257 const auto& index_array = mf.IndexArray();
258 if (index_array.size() == 1) {
259 Box const& b = strided_box(mf.box(index_array[0]), stride,
offset);
260 const IntVect lo = b.smallEnd();
264 j = lo[1] + (j-lo[1])*stride[1];,
265 k = lo[2] + (k-lo[2])*stride[2];)
270 constexpr int MT = 128;
272 constexpr int MT = AMREX_GPU_MAX_THREADS;
274 ParallelForStrided_doit<MT>(mf, stride,
offset, f);
278template <FabArrayType MF,
typename F>
280ParallelForRedBlack_doit (MF
const& mf,
int redblack,
F const& f)
282 const auto& index_array = mf.IndexArray();
283 if (index_array.size() == 1) {
285 Box b = mf.box(index_array[0]);
287 const int xhi = b.bigEnd(0);
292 const int i = 2*ic + ((j+k+redblack) & 1);
293 if (i >= xlo && i <= xhi) {
299 constexpr int MT = 128;
301 constexpr int MT = AMREX_GPU_MAX_THREADS;
303 ParallelForRedBlack_doit<MT>(mf, redblack, f);
#define AMREX_ASSERT(EX)
Definition AMReX_BLassert.H:38
#define AMREX_FORCE_INLINE
Definition AMReX_Extension.H:124
#define AMREX_GPU_ERROR_CHECK()
Definition AMReX_GpuError.H:151
#define AMREX_GPU_DEVICE
Definition AMReX_GpuQualifiers.H:18
#define AMREX_GPU_HOST_DEVICE
Definition AMReX_GpuQualifiers.H:20
Array4< int const > offset
Definition AMReX_HypreMLABecLap.cpp:1139
#define AMREX_D_TERM(a, b, c)
Definition AMReX_SPACE.H:172
virtual void * alloc(std::size_t sz)=0
Allocate sz bytes from this arena.
__host__ __device__ const IntVectND< dim > & smallEnd() const &noexcept
Return the inclusive lower bound of the box.
Definition AMReX_Box.H:124
__host__ __device__ constexpr Dim3 dim3() const noexcept
Definition AMReX_IntVect.H:262
amrex_long Long
Definition AMReX_INT.H:30
__host__ __device__ BoxND< dim > coarsen(const BoxND< dim > &b, int ref_ratio) noexcept
Return a copy of b coarsened by the isotropic ratio ref_ratio.
Definition AMReX_Box.H:1469
__host__ __device__ BoxND< dim > grow(const BoxND< dim > &b, int i) noexcept
Return a copy of b grown uniformly by i cells in every direction.
Definition AMReX_Box.H:1326
Arena * The_Pinned_Arena()
Definition AMReX_Arena.cpp:869
Arena * The_Arena()
Definition AMReX_Arena.cpp:829
void freeAsync(Arena *arena, void *mem) noexcept
Definition AMReX_GpuDevice.H:345
void htod_memcpy_async(void *p_d, const void *p_h, const std::size_t sz) noexcept
Definition AMReX_GpuDevice.H:421
gpuStream_t gpuStream() noexcept
Definition AMReX_GpuDevice.H:291
void ParallelFor(TypeList< CTOs... > ctos, std::array< int, sizeof...(CTOs)> const &runtime_options, T N, F &&f)
Definition AMReX_CTOParallelForImpl.H:202
BoxND< 3 > Box
Box is an alias for amrex::BoxND instantiated with AMREX_SPACEDIM.
Definition AMReX_BaseFwd.H:35
BoxIndexerND< 3 > BoxIndexer
Definition AMReX_Box.H:2608
IntVectND< 3 > IntVect
IntVect is an alias for amrex::IntVectND instantiated with AMREX_SPACEDIM.
Definition AMReX_BaseFwd.H:38
const int[]
Definition AMReX_BLProfiler.cpp:1665